/*ProfiPacket - packet radio terminal program
  Copyright (C) 1999  Alexander Feigl

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Author:

  Alexander Feigl
  Burachstrae 51

  D-88250 Weingarten

  Mail : Alexander.Feigl@gmx.de
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "TNC.h"
#include "Fork.h"
#include "TNCThread.h"

#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <termios.h>
#include <sys/select.h>
#include <sys/stat.h>

//
//
//
// TNC class : forking code + initalization
//
//
//


static int Fork_Helper(TNC *tptr);


TNC::TNC() 
 {

  inpacket_type=-1; 
  inpacket_size=-1; 
  inpacket_received=0;

  TNCFork=new Fork(Fork_Helper,this);

  connect(TNCFork,SIGNAL(receivedStdout(KProcess *,char *,int)),
          this,SLOT(receivedStdout(KProcess *,char *,int)) );

  TNCFork->start(KProcess::NotifyOnExit, (KProcess::Communication)
    (KProcess::Stdin|KProcess::Stdout));

  setInterface("/dev/ttyS0",0,0,9600);
  globalConstruct();
 }

TNC::~TNC() 
 {
  delete(TNCFork);
 }

static int Fork_Helper(TNC *tptr) 
 {
  fcntl(STDIN_FILENO,F_SETFL,O_NONBLOCK); 
  tptr->justForked(); 
  return(0);
 }


//
//
//
// TNC class : packet I/O
//
//
//

void TNC::putPacket(int type,int size,void *buffer)
 {
  TNCFork->writeStdin((char *)&type,sizeof(int));
  TNCFork->writeStdin((char *)&size,sizeof(int));
  if (size!=0) TNCFork->writeStdin((char *)buffer,size);
 }

void TNC::receivedStdout(KProcess *,char *buffer,int bufflen) 
 {
  int insize;

  while (1) 
   {
    if ( (inpacket_type==-1) || (inpacket_size==-1) ) 
     { 
      insize=sizeof(int);
     }
    else 
     {
      insize=inpacket_size;
     }

    while (  (inpacket_received<insize) && bufflen>0) {
      inpacket_buffer[inpacket_received]=*buffer;
      buffer++; 
      inpacket_received++;
      --bufflen;
     }

    if (inpacket_received<insize) return; 

    inpacket_received=0; 
    if (inpacket_type==-1)
     { 
      inpacket_type=*((int *) inpacket_buffer); 
      continue;
     }
    if (inpacket_size==-1) 
     {
      inpacket_size=*((int *) inpacket_buffer); 
      continue;
     }    
    receivePacket(inpacket_type,inpacket_size,inpacket_buffer);
    inpacket_size=-1; 
    inpacket_type=-1;
   }


 }


//
//
//
// TNCThread class : packet I/O
//
//
//

void TNCThread::putPacket(int type,int size,void *buffer)
 {
  fwrite(&type,sizeof(int),1,stdout);
  fwrite(&size,sizeof(int),1,stdout);
  if (size!=0) fwrite(buffer,1,size,stdout);  
  fflush(stdout);
 }

int TNCThread::getPacket(void)
 {
  int xchar;

  if ( (inpacket_type!=-1) && (inpacket_size!=-1) && 
       (inpacket_received==inpacket_size) ) 
   {
    inpacket_type=-1;
    inpacket_received=0;
   }

  if (inpacket_type==-1)
   {
    while ( inpacket_received!=sizeof(int) && ((xchar=getc(stdin))!=EOF) )
     {
      inpacket_buffer[inpacket_received]=xchar;
      ++inpacket_received;
     }
    if (inpacket_received!=sizeof(int)) return(0);
    inpacket_type=*((int *) inpacket_buffer);
    inpacket_received=0;
    inpacket_size=-1;
   }

  if (inpacket_size==-1)
   {
    while ( inpacket_received!=sizeof(int) && ((xchar=getc(stdin))!=EOF) )
     {
      inpacket_buffer[inpacket_received]=xchar;
      ++inpacket_received;
     }
    if (inpacket_received!=sizeof(int)) return(0);
    inpacket_size=*((int *) inpacket_buffer);
    inpacket_received=0;
   }

  while ( inpacket_received<inpacket_size && ((xchar=getc(stdin))!=EOF) )
   {
    inpacket_buffer[inpacket_received]=xchar;
    ++inpacket_received;
   }
  if (inpacket_received!=inpacket_size) return(0);
  return(1);
 }

//
//
//
// TNCThread class : main thread
//
//
//

void TNCThread::execThread(void)
 {
  struct fd_set myset;

  while (1)
   {
    FD_ZERO(&myset);
    if (interface_open)
     {
      FD_SET(SerialInterface,&myset);
     }
    FD_SET(STDIN_FILENO,&myset);
    select(999,&myset,0,0,0);    //FIXME : timeout for resync TNC
    while (getPacket())
     {
      receivedPacket(inpacket_type,inpacket_size,inpacket_buffer);
     }
    
   }
 }


int TNCThread::openInterface(char *device,int,int,int baudrate)
 {
  struct termios ttyset;
  struct stat lstat;
  int speed;

  switch (baudrate) 
   {
    case 50:
      speed=B50;
      break;
    case 75:
      speed=B75;
      break;
    case 110:
      speed=B110;
      break;
    case 134:
      speed=B134;
      break;
    case 150:
      speed=B150;
      break;
    case 200:
      speed=B200;
      break;
    case 300:
      speed=B300;
      break;
    case 600:
      speed=B600;
      break;
    case 1200:
      speed=B1200;
      break;
    case 1800:
      speed=B1800;
      break;
    case 2400:
      speed=B2400;
      break;
    case 4800:
      speed=B4800;
      break;
    case 9600:
      speed=B9600;
      break;
    case 19200:
      speed=B19200;
      break;
    case 38400:
      speed=B38400;
      break;
    case 57600: 
      speed=B57600;
      break;
    case 115200:
      speed=B115200;
      break;
    case 230400:
      speed=B230400;
      break;
    default:
      return(0);
   }
  if (interface_open) closeInterface();

  sprintf(lockname,"/var/lock/LCK...%s",device);
  if (stat(lockname,&lstat)==0)
   {
    return(0);
   }

  SerialInterface=open(device,O_RDWR|O_NONBLOCK);
  if (SerialInterface==-1)
   {
    return(0);
   }
  tcgetattr(SerialInterface,&ttyset);
  ttyset.c_iflag=IGNBRK|IGNPAR;
  ttyset.c_oflag=0;
  ttyset.c_cflag=CS8|CREAD|HUPCL|CLOCAL;
  ttyset.c_lflag=0;
  cfsetospeed(&ttyset,speed);
  cfsetispeed(&ttyset,speed);
  tcsetattr(SerialInterface,TCSANOW,&ttyset);
  interface_open=1; 

  //FIXME: UUCP locking doesn't work because of permissions
   {
    int i;
    i=creat(lockname,666);
    if (i!=-1)
     {
      close(i);
      have_lock=1;
     }
     else
     {
      have_lock=0;
     }
   }
  return(1);
 }

void TNCThread::closeInterface(void)
 {
  if (!interface_open) return;
  interface_open=0;
  close(SerialInterface);
  if (have_lock)
   {
    remove(lockname);
   }
 }
