/*ProfiPacket - packet radio terminal program
  Copyright (C) 1999  Alexander Feigl

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Author:

  Alexander Feigl
  Burachstrae 51

  D-88250 Weingarten

  Mail : Alexander.Feigl@gmx.de
*/

/*       ConvertMail0to1.c       Version 1.00

         Converts a version 0 database (PP Version <0.25) to the
         new version 1 standard. All files in forward queue will
         be lost due to changes in the forward queueing system.

         Functions of this programme :

         - convert the bbs.idx to new standard
         - deleting #?.fwd files, which are now longer needed

         Arguments:
      
         ConvertMail0to1 [BoxDataDir]

         Return Codes:

          0 - all is OK
         20 - wrong arguments
         21 - cannot open bbs.idx for read
         22 - cannot open bbs.new for write
         23 - cannot read the BBS index header
         24 - file is not a BBS index
         25 - file is not a version 0 BBS index
         26 - error while writing the BBS index header
         27 - error while writing the BBS box table
         28 - error while reading a mail from old index
         29 - error while writing a mail to new index
         30 - cannot delete old mailbox index
         31 - cannot rename new index to old 
*/


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <exec/memory.h>
#include <exec/ports.h>
#include <exec/devices.h>
#include <exec/io.h>
#include <exec/semaphores.h>
#include <dos/exall.h>
#include <utility/tagitem.h>

#include <clib/exec_protos.h>
#include <clib/dos_protos.h>

#include "Mail.h"


struct MailBase bbs_mailbase;
unsigned long   bbs_boxes[512];


int main(int argc,char **argv)
{
 unsigned char path[512];
 unsigned char fname[512];
 
 BPTR oldindex;
 BPTR newindex;

 if (argc!=2) return (20);
 strcpy(path,argv[1]);
 
 strcpy(fname,path);
 AddPart(fname,"bbs.idx",512);
 oldindex=Open(fname,MODE_OLDFILE);
 if (oldindex==NULL) return(21);
 strcpy(fname,path);
 AddPart(fname,"bbs.new",512);
 newindex=Open(fname,MODE_NEWFILE);
 if (newindex==NULL) 
  {
   Close(oldindex);
   return(22);
  }
 
 if (Read(oldindex,&bbs_mailbase,20)!=20)
  {
   Close(oldindex);
   Close(newindex);
   return(23);
  }
 
 if (strncmp(bbs_mailbase.IDString,"BBSINDEX",8)!=0)
  {
   Close(oldindex);
   Close(newindex);
   return(24);
  }

 if (bbs_mailbase.SizeofMail!=sizeof(struct MailOld))
  {
   Close(oldindex);
   Close(newindex);
   return(25);
  }

 bbs_mailbase.SizeofMail=sizeof(struct Mail);
 bbs_mailbase.IndexVersion=1;
  {
   unsigned long i;
   for (i=0;i<16;++i) bbs_mailbase.BoxIdent[i]='\0';
  }
 bbs_mailbase.WPUpdateDay=0;
  {
   unsigned long i;
   for (i=0;i<16;++i) bbs_mailbase.Padding[i]=0;
  }

 if (Write(newindex,&bbs_mailbase,sizeof(bbs_mailbase))!=
     sizeof(bbs_mailbase))
  {
   Close(oldindex);
   Close(newindex);
   return(26);
  }

  {
   unsigned long i;
   for (i=0;i<512;++i) bbs_boxes[i]=0;
  }

 if (Write(newindex,bbs_boxes,512*sizeof(unsigned long))!=
    512*sizeof(unsigned long))
  {
   Close(oldindex);
   Close(newindex);
   return(27);
  }
  {
   struct MailOld oldmail;
   struct Mail newmail;
   unsigned long MailNum;
  
   for (MailNum=0;MailNum<bbs_mailbase.NumberMails;++MailNum)
    {
     if (Read(oldindex,&oldmail,sizeof(oldmail))!=
         sizeof(oldmail))
      {
       Close(oldindex);
       Close(newindex);
       return(28);
      }

     /* Converting a mail */

     memset (&newmail,0,sizeof(newmail));
     newmail.Number=oldmail.Number;
     strcpy(newmail.Bid,oldmail.Bid);
     strcpy(newmail.From,oldmail.From);
     strcpy(newmail.To,oldmail.To);
     strcpy(newmail.At,oldmail.At);
     strcpy(newmail.Title,oldmail.Title);
     newmail.Date[0]=oldmail.Date[0];
     newmail.Date[1]=oldmail.Date[1];
     newmail.DateStatus[0]=oldmail.Date[0];
     newmail.DateStatus[1]=oldmail.Date[1];
     newmail.OriginDate[0]=oldmail.Date[0];
     newmail.OriginDate[1]=oldmail.Date[1];
     strcpy(newmail.OriginBBS,"DU1MMY");
     newmail.Status=oldmail.Status;
     newmail.Bytes=oldmail.Bytes;
     strcpy(newmail.HRoute,oldmail.HRoute);
     newmail.Lifetime=oldmail.Lifetime;


     if (Write(newindex,&newmail,sizeof(newmail))!=
         sizeof(newmail))
      {
       Close(oldindex);
       Close(newindex);
       return(29);
      }       
    }
  
  }
 Close(newindex);
 Close(oldindex);
 
  {
   BPTR xlock;
   xlock=Lock(path,SHARED_LOCK);
   if (xlock!=NULL)
    {
     struct FileInfoBlock *myfib;
     myfib=AllocDosObject(DOS_FIB,NULL);
     if (myfib!=NULL)
      {
       if (Examine(xlock,myfib)!=FALSE)
        {
         while (ExNext(xlock,myfib)!=FALSE)
          {
           unsigned char *xlast;
           xlast=myfib->fib_FileName;
           while (*(xlast)!='\0') xlast++;
           if ( (*(xlast-1)=='d') && (*(xlast-2)=='w') &&
                (*(xlast-3)=='f') && (*(xlast-4)=='.') )
            {
            strcpy(fname,path);
             AddPart(fname,myfib->fib_FileName,512);
             DeleteFile(fname);
            }
          }
        }
       FreeDosObject(DOS_FIB,myfib);
      }
     UnLock(xlock);
    }
  }
  {
   unsigned char fname2[512];
   strcpy(fname,path);
   AddPart(fname,"bbs.idx",512);
   strcpy(fname2,path);
   AddPart(fname2,"bbs.new",512);
   if (DeleteFile(fname)==FALSE) return (30);
   if (Rename(fname2,fname)==FALSE) return(31);
  }
 return(0);
}

