/*ProfiPacket - packet radio terminal program
  Copyright (C) 1999  Alexander Feigl

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Author:

  Alexander Feigl
  Burachstrae 51

  D-88250 Weingarten

  Mail : Alexander.Feigl@gmx.de
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#include <dos/dos.h>

#include <exec/memory.h>
#include <utility/tagitem.h>
#include <exec/ports.h>
#include <exec/semaphores.h>
#include <exec/io.h>
#include <dos/exall.h>

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <clib/exec_protos.h>
#include <clib/dos_protos.h>

#ifdef PRAGGI
#include <pragmas/exec_pragmas.h>
#include <pragmas/dos_pragmas.h>
#endif

#ifdef __GNUC_
#include <inline/exec.h>
#include <inline/dos.h>
#endif


#define IN_CONTRIB

#include "../User.h"

#include "../User.c"


struct version1header
  {
   char fileid[8];
   unsigned long fileversion;
   unsigned long usersize;
   unsigned long numusers;
  };

struct version2header
        {
         char fileid[8];
         unsigned long fileversion;
         unsigned long usersize;
         unsigned long userspace;
         unsigned long usedusers;
         unsigned char ownername[80];
         unsigned char ownerstreet[80];
         unsigned char ownercity[80];
         unsigned long password[4];
         unsigned long identifier;
        };


struct filestruct
        {
   unsigned long fileversion;
         unsigned long usersize;
         unsigned long userstart;
         unsigned long fileentries;
   unsigned char ownerfirstname[80];
   unsigned char ownerqth[80];
   unsigned char ownerlocator[80];
   unsigned long password[4];
   unsigned long identifier;
   unsigned long usedusers;
        };


int readheader(BPTR hfile,struct filestruct *finfo);
int writeheader(BPTR hfile,struct filestruct *finfo);
int writeidxheader(BPTR hfile,struct filestruct *finfo);
int readcuser(struct filestruct *finfo,BPTR ufile,struct User *luser,unsigned long i);
int callcheck(struct User *user);

 





int callcheck(struct User *cuser)
{
 unsigned char *xptr;
 unsigned long i,inssid;

 xptr=cuser->Call;

 if (strcmp("STANDARD",xptr)==0) return(TRUE);
 i=0;
 inssid=FALSE;
 while ( *xptr!='\0')
  {
   if (!inssid)
    {
     if (*xptr=='-')
      {
       inssid=TRUE;
       i=0;
      }
      else
      {
       ++i;
       if (i>6) return (FALSE);
       if ( ( (*xptr<'A') || (*xptr>'Z') ) &&
            ( (*xptr<'0') || (*xptr>'9') ) ) return (FALSE);
      }
    }
    else
    {
     ++i;
     if (i>2) return (FALSE);
     if ( (*xptr<'0') || (*xptr>'9')) return (FALSE);
    }
   ++xptr;
  }
 if (i==0) return(FALSE);
 return(TRUE);
}

int readcuser(struct filestruct *finfo,BPTR ufile,struct User *luser,unsigned long i)
{
 switch (finfo->fileversion)
  {
   case 1:
    {
     char *xptr;

     xptr=luser->Call;
     if (
            (Read(ufile,xptr,finfo->usersize))!=finfo->usersize) return (FALSE);
     if (i==0) strcpy(luser->Call,"STANDARD");
     return (TRUE);
    }
   case 2:
    {
     return (QSO_LoadUser(luser,ufile,finfo->userstart+finfo->usersize*i,finfo->usersize));
    }
   case 3:
    {
     return (QSO_LoadUser(luser,ufile,finfo->userstart+finfo->usersize*i,finfo->usersize));
    }
  }
 return(FALSE);
}

int writeidxheader(BPTR hfile,struct filestruct *finfo)
{
 struct version2indexhead v2ih;

 strcpy(v2ih.fileid,"USERINDX");
  {
   v2ih.fileversion=3;
   v2ih.indexentries=finfo->usedusers;
  }
 if ( (Seek(hfile,0,OFFSET_BEGINNING))==-1) return (FALSE);
 if ( (Write(hfile,&v2ih,sizeof(v2ih)))!=sizeof(v2ih)) return (FALSE);
 return(TRUE);

}

int writeheader(BPTR hfile,struct filestruct *finfo)
        {
         struct version3header v3h;

         memset (&v3h,0,sizeof(v3h));
         memcpy (v3h.fileid,"PACKUSER",8);
         v3h.fileversion=3;
         v3h.usersize=finfo->usersize;
         v3h.userspace=finfo->usedusers;
         v3h.usedusers=finfo->usedusers;
         strcpy (v3h.ownerfirstname,finfo->ownerfirstname);
         strcpy (v3h.ownerqth,finfo->ownerqth);
         strcpy (v3h.ownerlocator,finfo->ownerlocator);
         v3h.password[0]=finfo->password[0];
         v3h.password[1]=finfo->password[1];
         v3h.password[2]=finfo->password[2];
         v3h.password[3]=finfo->password[3];
         v3h.identifier=finfo->identifier;

   if ( (Seek(hfile,0,OFFSET_BEGINNING))==-1)
      return(FALSE);

   if ( (Write(hfile,&v3h,sizeof(v3h)))!=sizeof(v3h))
      return(FALSE);

   return(TRUE);
        }

int readheader(BPTR hfile,struct filestruct *finfo)
 {
  signed long filesize;
  if ( (Seek(hfile,0,OFFSET_BEGINNING))==-1)
  return(FALSE);
   {
    char fileheader[12];
    if ( (Read(hfile,fileheader,12))!=12)
       return(FALSE);
    if ( (fileheader[0]!='P') || (fileheader[1]!='A') ||
         (fileheader[2]!='C') || (fileheader[3]!='K') ||
         (fileheader[4]!='U') || (fileheader[5]!='S') ||
         (fileheader[6]!='E') || (fileheader[7]!='R') )
       return(FALSE);
    finfo->fileversion=*((unsigned long *) (fileheader+8) );
   }

 if ( (Seek(hfile,0,OFFSET_END))==-1)
       return(FALSE);

 if ( (filesize=Seek(hfile,0,OFFSET_BEGINNING))==-1)
 return(FALSE);

 switch (finfo->fileversion)
    {
     case 1:
      {
       struct version1header v1h;
       finfo->userstart=sizeof(struct version1header);
       if ( (Read(hfile,&v1h,finfo->userstart))!=finfo->userstart)
         return(FALSE);
       finfo->usersize=v1h.usersize;
       finfo->fileentries=(filesize - finfo->userstart) / finfo->usersize;
       finfo->ownerfirstname[0]='\0';
       finfo->ownerqth[0]='\0';
       finfo->ownerlocator[0]='\0';
       finfo->password[0]=0;
       finfo->password[1]=0;
       finfo->password[2]=0;
       finfo->password[3]=0;
       finfo->identifier=0;
       break;
      }

     case 2:
      {
       struct version2header v2h;
       finfo->userstart=sizeof(struct version2header);
       if ( (Read(hfile,&v2h,finfo->userstart))!=finfo->userstart)
         return(FALSE);
       finfo->usersize=v2h.usersize;
       finfo->fileentries=(filesize - finfo->userstart) / finfo->usersize;
       finfo->ownerfirstname[0]='\0';
       finfo->ownerqth[0]='\0';
       finfo->ownerlocator[0]='\0';
       finfo->password[0]=v2h.password[0];
       finfo->password[1]=v2h.password[1];
       finfo->password[2]=v2h.password[2];
       finfo->password[3]=v2h.password[3];
       finfo->identifier=v2h.identifier;
       break;
      }

     case 3:
      {
       struct version3header v3h;
       finfo->userstart=sizeof(struct version3header);
       if ( (Read(hfile,&v3h,finfo->userstart))!=finfo->userstart)
         return(FALSE);
       finfo->usersize=v3h.usersize;
       finfo->fileentries=(filesize - finfo->userstart) / finfo->usersize;
       strcpy(finfo->ownerfirstname,v3h.ownerfirstname);
       strcpy(finfo->ownerqth,v3h.ownerqth);
       strcpy(finfo->ownerlocator,v3h.ownerlocator);
       finfo->password[0]=v3h.password[0];
       finfo->password[1]=v3h.password[1];
       finfo->password[2]=v3h.password[2];
       finfo->password[3]=v3h.password[3];
       finfo->identifier=v3h.identifier;
       break;
      }


     default:
      return(FALSE);
    }
   finfo->usedusers=0;
   return(TRUE);
  }

int main(int argc,char **argv)
        {
         struct filestruct finfo;
         BPTR sourcefile,rawfile,indexfile;
         char sourcefn[256];
         char rawfn[256];
         char indexfn[256];
         unsigned long loffset,usernr;
         int stdwritten=FALSE;

   if (argc!=2)
    {
     printf ("Benutzung : %s [Dateiname der Benutzerdatei (users.dat)]\n",
            argv[0]);
     return(20);

    }

   puts("ProfiPacket_ConvertUsers V1.2\n");

   {
    char *ptr1,*ptr2;
    ptr1=NULL;
    ptr2=(char *) ((struct User *)ptr1)->Call;
    loffset=ptr2-ptr1;
   }

   strcpy (sourcefn,argv[1]);
   strcpy (rawfn,sourcefn);
   strcpy (indexfn,sourcefn);
        {
         char *sptr,*dptr;
         sptr=sourcefn;
         dptr=NULL;
         while (*sptr!='\0')
                {
                 if (*sptr=='.') dptr=sptr;
                 sptr++;
                }
         if (dptr==NULL)
                {
                 puts ("Dateiname ungltig!");
         return(20);
                }
         *(dptr+1)='b';
         *(dptr+2)='a';
         *(dptr+3)='k';
         *(dptr+4)='\0';
         dptr=dptr-sourcefn+indexfn;
         *(dptr+1)='i';
         *(dptr+2)='d';
         *(dptr+3)='x';
         *(dptr+4)='\0';
        }

   printf("Dateiname Originaldatei : %s\n",sourcefn);
   printf("Dateiname neue Datei    : %s\n",rawfn);
   printf("Dateiname Indexdatei    : %s\n\n",indexfn);

   DeleteFile(sourcefn);
   DeleteFile(indexfn);
   Rename(rawfn,sourcefn);

   sourcefile=Open(sourcefn,MODE_OLDFILE);



   if (sourcefile==NULL)
    {
       finfo.fileversion=0;
       finfo.userstart=0;
       finfo.usersize=0;
       finfo.fileentries=0;
       finfo.ownerfirstname[0]='\0';
       finfo.ownerqth[0]='\0';
       finfo.ownerlocator[0]='\0';
       finfo.password[0]=0;
       finfo.password[1]=0;
       finfo.password[2]=0;
       finfo.password[3]=0;
       finfo.identifier=0;
       finfo.usedusers=0;
    }


   if ((indexfile=Open(indexfn,MODE_NEWFILE))==NULL)
        {
         puts("Fehler : Kann Indexdatei nicht ffnen!");
         if (sourcefile) Close(sourcefile);
         return (20);
        }

   if ((rawfile=Open(rawfn,MODE_NEWFILE))==NULL)
        {
         puts("Fehler : Kann neue Datei nicht ffnen!");
         Close(indexfile);
         if (sourcefile) Close(sourcefile);
         DeleteFile(indexfn);
         return (20);
        }


   if (sourcefile)

    {
     if (!readheader(sourcefile,&finfo))
          {
           puts("Fehler : Datei defekt / keine Benutzerdatei!");
           Close(rawfile);
           Close(indexfile);
           if (sourcefile) Close(sourcefile);
           DeleteFile(indexfn);
           DeleteFile(rawfn);
           return(20);
          }
    }

   switch (finfo.fileversion)
        {
         case 0:
           puts("ProfiPacket Benutzerdatei noch nicht angelegt");
           break;
         case 1:
           puts("ProfiPacket Benutzerdatei Version 1 (bis V 0.19)");
           break;
         case 2:
           puts("ProfiPacket Benutzerdatei Version 2 (ab  V 0.20)");
           break;
         case 3:
           puts("ProfiPacket Benutzerdatei Version 3 (ab  V 0.25)");
        }

   printf ("Dateieintrge    : %ld\n\n",finfo.fileentries);

    {
     struct version2header v3h;
     if ( (Write(rawfile,&v3h,sizeof(v3h)))!=sizeof(v3h))
      {
       puts("\n\nFehler : Kann Header nicht schreiben!");
       Close(rawfile);
       Close(indexfile);
       if (sourcefile) Close(sourcefile);
       DeleteFile(rawfn);
       DeleteFile(indexfn);
      }
    }

    {
     struct version2indexhead v2h;
     if ( (Write(indexfile,&v2h,sizeof(v2h)))!=sizeof(v2h))
      {
       puts("\n\nFehler : Kann Header nicht schreiben!");
       Close(rawfile);
       Close(indexfile);
       if (sourcefile) Close(sourcefile);
       DeleteFile(rawfn);
       DeleteFile(indexfn);
      }

    }

   printf ("Konvertiere Benutzer ");
   fflush (stdout);

        {
         struct User tempuser;
         QSO_ResetUser(&tempuser);
         for (usernr=0;usernr<finfo.fileentries;++usernr)
                {
                 if (!readcuser(&finfo,sourcefile,&tempuser,usernr))
                  {
                   puts("\n\nFehler : Benutzerverzeichnis zerstrt!");
                   Close(rawfile);
                   Close(indexfile);
                   if (sourcefile) Close(sourcefile);
                   DeleteFile(rawfn);
                   DeleteFile(indexfn);
                   exit(20);
                  }

                 if (tempuser.Call[0]=='\0')
                  {
                   putchar('0');
                  }
                  else
                  {
                   if (!callcheck(&tempuser))
                    {
                     putchar('-');
                    }
                   else
                    {
                     if ( strcmp("STANDARD",tempuser.Call)==0) stdwritten=TRUE;

                     if (!QSO_StoreUser(&tempuser,rawfile,
                          sizeof(struct version3header)+
                          finfo.usedusers*(sizeof(struct User)-loffset)))
                      {
                       puts("\n\nFehler : Kann Benutzer nicht schreiben!");
                       Close(rawfile);
                       Close(indexfile);
                       if (sourcefile) Close(sourcefile);
                       DeleteFile(rawfn);
                       DeleteFile(indexfn);
                       exit(20);
                      }

                     {
                      struct version2index v2i;

                      strcpy (v2i.call,tempuser.Call);
                      v2i.filepos=finfo.usedusers;
                      if ( Write(indexfile,&v2i,sizeof(v2i))!=
                           sizeof(v2i))
                      {
                       puts("\n\nFehler : Kann Benutzerindex nicht schreiben!");
                       Close(rawfile);
                       Close(indexfile);
                       if (sourcefile) Close(sourcefile);
                       DeleteFile(rawfn);
                       DeleteFile(indexfn);
                       exit(20);
                      }

                     }
                    finfo.usedusers++;
                    putchar('.');
                   }


                  }
                  fflush(stdout);

                }
        }


   if (!stdwritten)
    {
     struct User tempuser;
     QSO_ResetUser(&tempuser);
     strcpy (tempuser.Call,"STANDARD");
     if (!QSO_StoreUser(&tempuser,rawfile,
        sizeof(struct version3header)+
        finfo.usedusers*(sizeof(struct User)-loffset)))
       {
        puts("\n\nFehler : Kann Benutzer nicht schreiben!");
        Close(rawfile);
        Close(indexfile);
        if (sourcefile) Close(sourcefile);
        DeleteFile(rawfn);
        DeleteFile(indexfn);
       }

      {
       struct version2index v2i;
       strcpy (v2i.call,tempuser.Call);
       v2i.filepos=finfo.usedusers;
       if ( Write(indexfile,&v2i,sizeof(v2i))!=
            sizeof(v2i))
        {
         puts("\n\nFehler : Kann Benutzerindex nicht schreiben!");
         Close(rawfile);
         Close(indexfile);
         if (sourcefile) Close(sourcefile);
         DeleteFile(rawfn);
         DeleteFile(indexfn);
        }

      }
     finfo.usedusers++;
     puts("\n\nStandardbenutzer hinzugefgt");

    }
   finfo.usersize=sizeof(struct User)-loffset;

   printf ("\n\nSchreibe neuen Header ...\n");
   if (!writeheader(rawfile,&finfo))
        {
         puts("\n\nFehler : Kann Header nicht schreiben!");
         Close(rawfile);
         Close(indexfile);
         if (sourcefile) Close(sourcefile);
         DeleteFile(rawfn);
         DeleteFile(indexfn);
         return(20);
        }

   if (!writeidxheader(indexfile,&finfo))
        {
         puts("\n\nFehler : Kann Index-Header nicht schreiben!");
         Close(rawfile);
         Close(indexfile);
         if (sourcefile) Close(sourcefile);
         DeleteFile(rawfn);
         DeleteFile(indexfn);
         return(20);
        }


   if (sourcefile) Close(sourcefile);
   Close(rawfile);
   Close(indexfile);

   return(0);


 }

